@@ -2,6 +2,7 @@
 # this repository contains the full copyright notices and license terms.
 
 from trytond.pool import Pool
+
 from . import stock
 
 
@@ -15,6 +16,3 @@ def register():
         stock.ShipmentOut,
         stock.ShipmentInReturn,
         module='stock_package', type_='model')
-    Pool.register(
-        stock.PackageLabel,
-        module='stock_package', type_='report')
