@@ -1,17 +1,16 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
 from sql import Cast, Literal
-from sql.functions import Substring, Position
+from sql.functions import CharLength, Position, Substring
 
-from trytond import backend
 from trytond.i18n import gettext
-from trytond.model import ModelSQL, ModelView, Workflow, fields, tree
+from trytond.model import (
+    DeactivableMixin, ModelSQL, ModelView, Workflow, fields, tree)
 from trytond.pool import Pool, PoolMeta
-from trytond.pyson import Eval, Bool, Id, If
-from trytond.report import Report
+from trytond.pyson import Bool, Eval, Id, If
 from trytond.transaction import Transaction
 
-from .exceptions import PackageError
+from .exceptions import PackageError, PackageValidationError
 
 
 class Configuration(metaclass=PoolMeta):
@@ -19,7 +18,8 @@ class Configuration(metaclass=PoolMeta):
     package_sequence = fields.MultiValue(fields.Many2One(
             'ir.sequence', "Package Sequence", required=True,
             domain=[
-                ('company', 'in', [Eval('context', {}).get('company'), None]),
+                ('company', 'in', [
+                        Eval('context', {}).get('company', -1), None]),
                 ('sequence_type', '=',
                     Id('stock_package', 'sequence_type_package')),
                 ]))
@@ -45,27 +45,7 @@ class ConfigurationSequence(metaclass=PoolMeta):
             ('company', 'in', [Eval('company', -1), None]),
             ('sequence_type', '=',
                 Id('stock_package', 'sequence_type_package')),
-            ],
-        depends=['company'])
-
-    @classmethod
-    def __register__(cls, module_name):
-        exist = backend.TableHandler.table_exist(cls._table)
-        if exist:
-            table = cls.__table_handler__(module_name)
-            exist &= table.column_exist('package_sequence')
-
-        super(ConfigurationSequence, cls).__register__(module_name)
-
-        if not exist:
-            cls._migrate_property([], [], [])
-
-    @classmethod
-    def _migrate_property(cls, field_names, value_names, fields):
-        field_names.append('package_sequence')
-        value_names.append('package_sequence')
-        super(ConfigurationSequence, cls)._migrate_property(
-            field_names, value_names, fields)
+            ])
 
     @classmethod
     def default_package_sequence(cls):
@@ -77,52 +57,127 @@ class ConfigurationSequence(metaclass=PoolMeta):
             return None
 
 
-class WeightMixin:
+class MeasurementsMixin:
     __slots__ = ()
 
+    length = fields.Float(
+        "Length", digits='length_uom',
+        help="The length of the package.")
+    length_uom = fields.Many2One(
+        'product.uom', "Length UoM",
+        domain=[('category', '=', Id('product', 'uom_cat_length'))],
+        states={
+            'required': Bool(Eval('length')),
+            },
+        help="The Unit of Measure for the package length.")
+    height = fields.Float(
+        "Height", digits='height_uom',
+        help="The height of the package.")
+    height_uom = fields.Many2One(
+        'product.uom', "Height UoM",
+        domain=[('category', '=', Id('product', 'uom_cat_length'))],
+        states={
+            'required': Bool(Eval('height')),
+            },
+        help="The Unit of Measure for the package height.")
+    width = fields.Float(
+        "Width", digits='width_uom',
+        help="The width of the package.")
+    width_uom = fields.Many2One(
+        'product.uom', "Width UoM",
+        domain=[('category', '=', Id('product', 'uom_cat_length'))],
+        states={
+            'required': Bool(Eval('width')),
+            },
+        help="The Unit of Measure for the package width.")
+
+    packaging_volume = fields.Float(
+        "Packaging Volume", digits='packaging_volume_uom',
+        states={
+            'readonly': (
+                Bool(Eval('length'))
+                & Bool(Eval('height'))
+                & Bool(Eval('width'))),
+            },
+        help="The volume of the package.")
+    packaging_volume_uom = fields.Many2One(
+        'product.uom', "Packaging Volume UoM",
+        domain=[('category', '=', Id('product', 'uom_cat_volume'))],
+        states={
+            'required': Bool(Eval('packaging_volume')),
+            },
+        help="The Unit of Measure for the packaging volume.")
+
     packaging_weight = fields.Float(
-        "Packaging Weight", digits=(16, Eval('packaging_weight_digits', 2)),
-        depends=['packaging_weight_digits'],
+        "Packaging Weight", digits='packaging_weight_uom',
         help="The weight of the package when empty.")
     packaging_weight_uom = fields.Many2One(
-        'product.uom', "Packaging Weight Uom",
+        'product.uom', "Packaging Weight UoM",
         domain=[('category', '=', Id('product', 'uom_cat_weight'))],
         states={
             'required': Bool(Eval('packaging_weight')),
             },
-        depends=['packaging_weight'])
-    packaging_weight_digits = fields.Function(
-        fields.Integer("Packaging Weight Digits"),
-        'on_change_with_packaging_weight_digits')
+        help="The Unit of Measure for the packaging weight.")
 
-    @fields.depends('packaging_weight_uom')
-    def on_change_with_packaging_weight_digits(self, name=None):
-        return (
-            self.packaging_weight_uom.digits if self.packaging_weight_uom
-            else None)
+    @fields.depends(
+        'packaging_volume', 'packaging_volume_uom',
+        'length', 'length_uom',
+        'height', 'height_uom',
+        'width', 'width_uom')
+    def on_change_with_packaging_volume(self):
+        pool = Pool()
+        ModelData = pool.get('ir.model.data')
+        Uom = pool.get('product.uom')
+
+        if not all([self.packaging_volume_uom,
+                    self.length, self.length_uom,
+                    self.height, self.height_uom,
+                    self.width, self.width_uom]):
+            if all([
+                        self.length,
+                        self.height,
+                        self.width]):
+                return
+            return self.packaging_volume
+
+        meter = Uom(ModelData.get_id('product', 'uom_meter'))
+        cubic_meter = Uom(ModelData.get_id('product', 'uom_cubic_meter'))
+
+        length = Uom.compute_qty(
+            self.length_uom, self.length, meter, round=False)
+        height = Uom.compute_qty(
+            self.height_uom, self.height, meter, round=False)
+        width = Uom.compute_qty(
+            self.width_uom, self.width, meter, round=False)
+
+        return Uom.compute_qty(
+            cubic_meter, length * height * width, self.packaging_volume_uom)
 
 
-class Package(tree(), WeightMixin, ModelSQL, ModelView):
+class Package(tree(), MeasurementsMixin, ModelSQL, ModelView):
     'Stock Package'
     __name__ = 'stock.package'
-    _rec_name = 'code'
-    code = fields.Char('Code', select=True, readonly=True, required=True)
+    _rec_name = 'number'
+    number = fields.Char("Number", readonly=True, required=True)
     company = fields.Many2One('company.company', "Company", required=True)
     type = fields.Many2One(
         'stock.package.type', "Type", required=True,
         states={
             'readonly': Eval('state') == 'closed',
-            },
-        depends=['state'])
+            })
     shipment = fields.Reference(
-        "Shipment", selection='get_shipment', select=True,
+        "Shipment", selection='get_shipment',
         states={
             'readonly': Eval('state') == 'closed',
             },
-        domain=[
-            ('company', '=', Eval('company', -1)),
-            ],
-        depends=['state', 'company'])
+        domain={
+            'stock.shipment.out': [
+                ('company', '=', Eval('company', -1)),
+                ],
+            'stock.shipment.in.return': [
+                ('company', '=', Eval('company', -1)),
+                ],
+            })
     moves = fields.One2Many('stock.move', 'package', 'Moves',
         domain=[
             ('company', '=', Eval('company', -1)),
@@ -130,23 +185,24 @@ class Package(tree(), WeightMixin, ModelSQL, ModelView):
             ('to_location.type', 'in', ['customer', 'supplier']),
             ('state', '!=', 'cancelled'),
             ],
+        filter=[
+            ('quantity', '!=', 0),
+            ],
         add_remove=[
             ('package', '=', None),
             ],
         states={
             'readonly': Eval('state') == 'closed',
-            },
-        depends=['company', 'shipment', 'state'])
+            })
     parent = fields.Many2One(
-        'stock.package', "Parent", select=True, ondelete='CASCADE',
+        'stock.package', "Parent", ondelete='CASCADE',
         domain=[
             ('company', '=', Eval('company', -1)),
             ('shipment', '=', Eval('shipment')),
             ],
         states={
             'readonly': Eval('state') == 'closed',
-            },
-        depends=['company', 'shipment', 'state'])
+            })
     children = fields.One2Many(
         'stock.package', 'parent', 'Children',
         domain=[
@@ -155,13 +211,17 @@ class Package(tree(), WeightMixin, ModelSQL, ModelView):
             ],
         states={
             'readonly': Eval('state') == 'closed',
-            },
-        depends=['company', 'shipment', 'state'])
+            })
     state = fields.Function(fields.Selection([
                 ('open', "Open"),
                 ('closed', "Closed"),
                 ], "State"), 'on_change_with_state')
 
+    @classmethod
+    def __setup__(cls):
+        cls.number.search_unaccented = False
+        super().__setup__()
+
     @classmethod
     def __register__(cls, module):
         pool = Pool()
@@ -170,6 +230,10 @@ class Package(tree(), WeightMixin, ModelSQL, ModelView):
         company_exist = table_h.column_exist('company')
         cursor = Transaction().connection.cursor()
 
+        # Migration from 6.8: rename code to number
+        if table_h.column_exist('code'):
+            table_h.column_rename('code', 'number')
+
         super().__register__(module)
 
         # Migration from 5.8: add company
@@ -188,6 +252,11 @@ class Package(tree(), WeightMixin, ModelSQL, ModelView):
                         [table.company], [value],
                         where=table.shipment.like(name + ',%')))
 
+    @classmethod
+    def order_number(cls, tables):
+        table, _ = tables[None]
+        return [CharLength(table.number), table.number]
+
     @classmethod
     def default_company(cls):
         return Transaction().context.get('company')
@@ -218,8 +287,37 @@ class Package(tree(), WeightMixin, ModelSQL, ModelView):
     @fields.depends('type')
     def on_change_type(self):
         if self.type:
-            self.packaging_weight = self.type.packaging_weight
-            self.packaging_weight_uom = self.type.packaging_weight_uom
+            for name in dir(MeasurementsMixin):
+                if isinstance(getattr(MeasurementsMixin, name), fields.Field):
+                    setattr(self, name, getattr(self.type, name))
+
+    @classmethod
+    def validate(cls, packages):
+        super().validate(packages)
+        for package in packages:
+            package.check_volume()
+
+    def check_volume(self):
+        pool = Pool()
+        Uom = pool.get('product.uom')
+        Lang = pool.get('ir.lang')
+        lang = Lang.get()
+        if not self.packaging_volume:
+            return
+        children_volume = 0
+        for child in self.children:
+            if child.packaging_volume:
+                children_volume += Uom.compute_qty(
+                    child.packaging_volume_uom, child.packaging_volume,
+                    self.packaging_volume_uom, round=False)
+        if self.packaging_volume < children_volume:
+            raise PackageValidationError(
+                gettext('stock_package.msg_package_volume_too_small',
+                    package=self.rec_name,
+                    volume=lang.format_number_symbol(
+                        self.packaging_volume, self.packaging_volume_uom),
+                    children_volume=lang.format_number_symbol(
+                        children_volume, self.packaging_volume_uom)))
 
     @classmethod
     def create(cls, vlist):
@@ -230,7 +328,7 @@ class Package(tree(), WeightMixin, ModelSQL, ModelView):
         config = Config(1)
         default_company = cls.default_company()
         for values in vlist:
-            values['code'] = config.get_multivalue(
+            values['number'] = config.get_multivalue(
                 'package_sequence',
                 company=values.get('company', default_company)).get()
         return super(Package, cls).create(vlist)
@@ -245,7 +343,7 @@ class Package(tree(), WeightMixin, ModelSQL, ModelView):
         return super().copy(packages, default=default)
 
 
-class Type(WeightMixin, ModelSQL, ModelView):
+class Type(MeasurementsMixin, DeactivableMixin, ModelSQL, ModelView):
     'Stock Package Type'
     __name__ = 'stock.package.type'
     name = fields.Char('Name', required=True)
@@ -254,14 +352,10 @@ class Type(WeightMixin, ModelSQL, ModelView):
 class Move(metaclass=PoolMeta):
     __name__ = 'stock.move'
     package = fields.Many2One(
-        'stock.package', "Package", select=True,
+        'stock.package', "Package", readonly=True,
         domain=[
             ('company', '=', Eval('company', -1)),
-            ],
-        states={
-            'readonly': Eval('state') == 'cancelled',
-            },
-        depends=['company', 'state'])
+            ])
 
     @property
     def package_path(self):
@@ -293,29 +387,17 @@ class Move(metaclass=PoolMeta):
 class PackageMixin(object):
     __slots__ = ()
     packages = fields.One2Many('stock.package', 'shipment', 'Packages',
+        domain=[
+            ('company', '=', Eval('company', -1)),
+            ])
+    root_packages = fields.One2Many('stock.package',
+        'shipment', 'Packages',
         domain=[
             ('company', '=', Eval('company', -1)),
             ],
-        depends=['company'])
-    root_packages = fields.Function(fields.One2Many('stock.package',
-            'shipment', 'Packages',
-            domain=[
-                ('company', '=', Eval('company', -1)),
-                ('parent', '=', None),
-                ],
-            depends=['company']),
-        'get_root_packages', setter='set_root_packages')
-
-    def get_root_packages(self, name):
-        return [p.id for p in self.packages if not p.parent]
-
-    @classmethod
-    def set_root_packages(cls, shipments, name, value):
-        if not value:
-            return
-        cls.write(shipments, {
-                'packages': value,
-                })
+        filter=[
+            ('parent', '=', None),
+            ])
 
     @classmethod
     def check_packages(cls, shipments):
@@ -323,7 +405,7 @@ class PackageMixin(object):
             if not shipment.packages:
                 continue
             length = sum(len(p.moves) for p in shipment.packages)
-            if len(list(shipment.packages_moves)) != length:
+            if len(shipment.packages_moves) != length:
                 raise PackageError(
                     gettext('stock_package.msg_package_mismatch',
                         shipment=shipment.rec_name))
@@ -350,10 +432,8 @@ class ShipmentOut(PackageMixin, object, metaclass=PoolMeta):
             Eval('warehouse_storage') == Eval('warehouse_output'),
             Eval('state') != 'waiting',
             Eval('state') != 'picked')
-        packages_depends = ['warehouse_storage', 'warehouse_output', 'state']
         for field in [cls.packages, cls.root_packages]:
             field.states['readonly'] = packages_readonly
-            field.depends.extend(packages_depends)
 
     @classmethod
     @ModelView.button
@@ -371,7 +451,9 @@ class ShipmentOut(PackageMixin, object, metaclass=PoolMeta):
 
     @property
     def packages_moves(self):
-        return (m for m in self.outgoing_moves if m.state != 'cancelled')
+        return [
+            m for m in self.outgoing_moves
+            if m.state != 'cancelled' and m.quantity]
 
     def _group_parcel_key(self, lines, line):
         try:
@@ -404,10 +486,8 @@ class ShipmentInReturn(PackageMixin, object, metaclass=PoolMeta):
     def __setup__(cls):
         super().__setup__()
         packages_readonly = ~Eval('state').in_(['waiting', 'assigned'])
-        packages_depends = ['state']
         for field in [cls.packages, cls.root_packages]:
             field.states['readonly'] = packages_readonly
-            field.depends.extend(packages_depends)
 
     @classmethod
     @ModelView.button
@@ -418,9 +498,6 @@ class ShipmentInReturn(PackageMixin, object, metaclass=PoolMeta):
 
     @property
     def packages_moves(self):
-        return (m for m in self.moves if m.state != 'cancelled')
-
-
-class PackageLabel(Report):
-    'Package Label'
-    __name__ = 'stock.package.label'
+        return [
+            m for m in self.moves
+            if m.state != 'cancelled' and m.quantity]
